/* Copyright (C) 2005, 2006, 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package compbase;

import memory.BootableImage;
import memory.MemoryImage;
import frontend.FrontAdmin;
import backend.ArchFactory;
import backend.Architecture;
import compression.CompressionFactory;
import compression.Compressor;
import osio.*;
import output.OutputFactory;
import output.OutputFormat;
import real.EmulReal;
import real.NativeReal;
import debug.DebugFactory;
import debug.DebugWriter;
import symbols.SymbolFactory;
import symbols.SymbolInformer;

/**
 * Context: root of all information needed for a compilation
 * 
 * @author S. Frenz
 * @version 090719 added support for complex written write state setting
 *  version 090717 added getUnwrittenVarList with associated methods and variables
 *  version 090626 added -x option to support stack extreme checks
 *  version 090508 removed -Z option as flow analysis is completely integrated in resolving
 *  version 090506 added -H option to support file listing as compiled, renamed insertReturnMissing into codeReturnMissing
 *  version 090505 added -Z option to disable newly inserted flow analysis
 *  version 090430 added -R option to support "return missing" exception
 *  version 090409 added compressedArchSpecialValue
 *  version 090407 added error check after Architecture.init
 *  version 090311 adopted removed reference to System and added osio-support for getTimeInfo
 *  version 090306 added -N option to support other entry methods than kernel.Kernel.main()
 *  version 090303 adopted changed osio package structure
 *  version 090301 got allocateMultiArray from MemoryImage
 *  version 090227 fixed release info and added version info
 *  version 090218 added support for synchronized-environment
 *  version 090207 added copyright notice
 *  version 081230 added call to Architecture.printStatistic
 *  version 081023 added verbose output "all done"
 *  version 081016 added -W option to support printing of method's instruction list globally
 *  version 080820 better help for invalid debug info writer parameter
 *  version 080714 reduced visability of internal writeSymInfo
 *  version 080712 added -D option to support different debug writers
 *  version 080706 added -K option to support alignment of object blocks
 *  version 080701 adopted changed symInfo-debug-interface
 *  version 080629 added -k option to suppress code generation of inline-marked methods
 *  version 080622 added support for static class initialization
 *  version 080617 changed output of kb->b for values <32kb
 *  version 080616 added -f option to skip generating throw-frames (not conforming Java!)
 *  version 080614 fixed parseInt for empty strings
 *  version 080613 added dataBlocks
 *  version 080612 removed parameter -f (force long-div/-mod through rte, replaced by arch-par -T rtlc), updated modes in syminfo
 *  version 080604 added support for basic exception environment
 *  version 080520 changed rteDAariLongCallMd to rteDAAriCallMd[]
 *  version 080517 adopted changes of Architecture.ariLongCall to ariCall
 *  version 080419 fixed typo in help text
 *  version 080202 added option for alternate newInstance and option to read options from a file
 *  version 080119 added runtime method to assign array pointer
 *  version 080118 fixed error output for unknown SymbolInformer
 *  version 080115 adopted changes in SymbolInformer.generateSymbols
 *  version 080104 added support for profiling and for -F to enable global profiling
 *  version 071214 added rteSIMnext to support chained interface maps
 *  version 070918 added error check after generating constant objects
 *  version 070815 added -T for architecture parameters
 *  version 070812 removed -F option as there are no plans to implement all float/double rte calls
 *  version 070808 changed -T to -V, added options -F for float-rte and -Y for native float
 *  version 070729 removed call to no longer needed SymbolInformer.prepareUnits
 *  version 070728 added support fur null-Strings in allocateString
 *  version 070727 beautify of syminfo output
 *  version 070714 added flags to generate all units/mthds, added allocateString
 *  version 070714 added output of size needed for in-image symbol information
 *  version 070711 added support for in-image symbol information
 *  version 070705 modified ramInitSize to respect allocClearBits
 *  version 070704 added embConstRAM
 *  version 070703 added constMemorySize
 *  version 070628 added err-checks
 *  version 070617 fixed marking SMthdBlock as used if not in leanRTE mode
 *  version 070615 added addrList
 *  version 070531 clarified statistic output
 *  version 070528 added string statistics
 *  version 070526 removed rteSArray* short-paths
 *  version 070522 fixed allocation of interface maps 
 *  version 070511 inserted step to create constant objects
 *  version 070506 added support for runtime environment without objects
 *  version 070504 splitted and renamed assignOffset to match real function
 *  version 070429 added output of compiler options to symbol information
 *  version 070331 fixed typos in output messages
 *  version 070312 added support for time measurement
 *  version 070303 added option for indirect accessed (movable) scalars
 *  version 070127 fixed embedded message
 *  version 070113 added rteDRIsArrayMd
 *  version 070105 removed image prefixing
 *  version 070101 fixed help, prevention of output for compressed temporary image, removed method-idx
 *  version 061228 added option for automatic method inlining and bound check skipping
 *  version 061225 added option to regulate method inlining
 *  version 061222 replaced streamlineHeader by !standardHeader
 *  version 061219 more helpful error on invalid parameters
 *  version 061217 added streamlineHeader option
 *  version 061211 added verbose mode, got debugCode from Architecture
 *  version 061129 took static objects from TypeRef and added ContextConfig
 *  version 061128 added support for multiplier in parseInt
 *  version 061127 added embedded option -e
 *  version 061111 minor changes
 *  version 061102 added option to specify location of struct-pointer
 *  version 061027 added dummy for structClass
 *  version 060908 full integration of bootstrap mode
 *  version 060818 changed options for -A parameter
 *  version 060817 added support for image compression
 *  version 060808 changed hierarchy to support CommBase
 *  version 060807 added support for runtime-call on div/mod-long-arithmetic operation
 *  version 060807 added support for byte-strings
 *  version 060720 added compilation support
 *  version 060714 added support for call instead of native bound/null exception
 *  version 060629 added support for call instead of ptr-assign
 *  version 060628 added support for static compilation
 *  version 060607 initial version
 */

public class Context {
  public TextPrinter out;
  public OsIO osio;
  
  public OutputFormat imgOut;
  public SymbolInformer symGen;
  public FrontAdmin fa;
  public Architecture arch;
  public MemoryImage mem;
  public BootableImage bootMem; //if mem instanceof BootableImage, then bootMem==mem
  public boolean embedded, embConstRAM, leanRTE, dynaMem, indirScalars, noThrowFrames, noInlineMthdObj;
  public boolean verbose, debugCode, printCode, err;
  public int ramStart, ramSize, ramInitAddr; //used only in embedded mode
  public int constMemorySize, symGenSize; //valid after genConstObj / generateSymbols
  public int maxStmtAutoInline;
  public int alignBlockMask; //initialized to 0
  public String debugPrefix;
  public boolean assignCall, assignHeapCall;
  public boolean doBoundCheck, runtimeBound, doArrayStoreCheck, runtimeNull;
  public boolean byteString, structAsReference;
  public boolean genAllUnitDesc, genAllMthds;
  public boolean globalProfiling, profilerUsed, globalStackExtreme, stackExtremeUsed, throwUsed, syncUsed;
  public boolean alternateObjNew;
  public boolean codeReturnMissing, returnMissingNativeNotRTE;
  
  public Pack root, rte;
  public Pack defUnits;
  public QualID langRoot;
  public String startMethod;
  public Unit langString;
  public Unit rteSArray, rteSClassDesc, rteSIntfDesc, rteSIntfMap, rteSMthdBlock;
  public Unit rteDynamicRuntime, rteDynamicAri, structClass;
  public Unit excThrowable, excChecked;
  public int rteSIMowner, rteSIMnext; //index in rteSIntfMap
  public Mthd rteDRNewInstMd, rteDRNewArrayMd, rteDRNewMultArrayMd;
  public Mthd rteDRIsInstMd, rteDRIsImplMd, rteDRIsArrayMd, rteDRCheckArrayStoreMd;
  public Mthd rteDRAssignMd, rteDRBoundExcMd, rteDRNullExcMd;
  public Mthd[] rteDAAriCallMds=new Mthd[StdTypes.MAXBASETYPE];
  public Mthd rteDRProfilerMd, rteDRStackExtremeErrMd; //rte-methods for profiler and stack extreme check
  public Mthd rteDoThrowMd; //method called if something is thrown
  public Mthd rteDoSyncMd; //method called if something is synchronized
  public Mthd rteRetMissMd; //method called if return is missing
  public Mthd staticInitMthds; //list of methods for static initialization
  public boolean staticInitDone; //flag to check if static initialization code is inserted somewhere
  public int rteThrowFrameOff; //offset/address of variable containing address of throw-frame
  public int rteStackExtremeOff; //offset/address of variable containg stack extreme value
  public int langStrVal, langStrCnt; //index/offset in langString;
  public int rteSClassParent, rteSClassImpl; //index in rteSClassDesc
  public int rteSClassInstScalarSize, rteSClassInstRelocTableEntries, rteSClassInstIndirScalarSize; //only for alternateNewObj
  public int rteSArrayLength, rteSArrayDim, rteSArrayStd, rteSArrayExt; //offsets/index in rteSArray
  public int indirScalarSizeOff, indirScalarAddrOff;
  public DebugWriter debugWriter;
  
  public TypeRef clssType, stringType, objectType;
  public CtxBasedConfig config;
  public AddrList freeAddrLists;
  
  public int stringCount, stringChars, stringMemBytes;
  public int mthdCount, mthdCodeSize;
  
  private UnitList unitList, lastUnit;
  public BootableImage compressedImage;
  public int compressedArchSpecialValue;
  private int compressedImageOrigLen;
  private byte[] compressedImageInfo;
  private StringList fileList;
  private Context decompressor;
  public DataBlockList dataBlocks;
  
  private boolean streamline, standardHeader=true, timing, prefereNativeReal;
  private int inlineLevels=3, memStart=-1, memSize=131072; //init with 128K image size starting at 1MB (entered below)
  private String headerFile;
  private byte[] headerData;
  private String[] secArgv;
  private String compressorName;
  private Compressor compressor;
  private String compiledFiledListname;
  private VrblList emptyVrblLists;
  
  public Context(OsIO iOS) {
    out=(osio=iOS).getNewFilePrinter(null);
    debugPrefix="";
    doBoundCheck=true;
    doArrayStoreCheck=true;
    startMethod="kernel.Kernel.main";
  }
  
  public int compile(String[] argv, String versionInfo) {
    UnitList ulist;
    boolean error=false;
    int i, uCnt=0;
    long t0=0l, t1=0l, t2=0l, t3=0l, t4=0l, t5=0l, t6=0l, t7=0l, t8=0l, t9=0l, t10=0l;
    
    //get output and check parameters
    if (compressedImage==null) {
      out.print("Welcome to the smfCompiler SJC");
      if (versionInfo!=null) {
        out.print(" (");
        out.print(versionInfo);
        out.print(')');
      }
      out.println();
    }
    else out.println("Compiling decompressor...");
    if (argv.length<1) {
      printHelp();
      return -1;
    }
    
    if (!getParameters(argv)) return -1;
    
    if (timing) {
      out.print("(t0) ");
      t0=osio.getTimeInfo();
    }
    //initialize environment
    if (arch==null) arch=ArchFactory.getArchitecture(null);
    if (memStart==-1) memStart=1048576; //set default value
    if (arch.relocBytes==2) {
      if (memStart>=0x10000) {
        out.println("Invalid base address for 16 bit architecture");
        return -1;
      }
      if (memStart+memSize>=0x10000) {
        memSize=0x10000-memStart;
        out.print("Memory for 16 bit architecture limited to ");
        out.print(memSize>>>10);
        out.println(" kb to fit address range");
      }
    }
    if (compressor!=null) {
      if (imgOut!=null) {
        out.println("Output format can not be specified for compressed image");
        return -1;
      }
      mem=bootMem=new BootableImage();
    }
    else if (memSize!=0) {
      if (imgOut==null) imgOut=OutputFactory.getOutputFormat(null);
      if (!imgOut.checkParameter(osio, out)) return -1;
      mem=bootMem=new BootableImage();
    }
    else {
      if (mem==null) {
        out.println("Requested bootstrap mode with mem==null");
        return -1;
      }
      if (imgOut!=null) {
        out.println("Requested bootstrap mode with imgOut!=null");
        return -1;
      }
    }
    arch.init(mem, inlineLevels, this);
    if (err || arch.err) {
      out.println("There were errors");
      return -1;
    }
    if (bootMem!=null) {
      if (!standardHeader && headerFile!=null && (headerData=osio.readFile(headerFile))==null) {
        out.print("Could not read header file ");
        out.println(headerFile);
      }
      bootMem.init(memSize, memStart, standardHeader, headerData, this);
    }
    else if (!standardHeader) {
      out.println("Bootstrap mode needs standard header");
      return -1;
    }
    mem.streamObjects=streamline;
    mem.noSizeScalars=alternateObjNew;
    if (arch.real==null) { //architecture may have initialized arch.real already
      if (prefereNativeReal) arch.real=new NativeReal();
      else arch.real=new EmulReal();
    }
    fa=new FrontAdmin(this);
    if (compiledFiledListname!=null) fa.activateListCompiledFiles(compiledFiledListname);
    
    //tell the current configuration
    if (verbose) printInfo();
    out.print("Compiling ");
    out.print(dynaMem ? "movable " : (embedded ? "embedded " : "static "));
    out.print(streamline ? "streamlined" : "fullsized");
    out.print(" objects for ");
    out.print(arch.relocBytes<<3);
    out.print(" bit, assign ");
    out.println(assignCall ? "all pointers via call"
        : (assignHeapCall ? "heap pointers via call" : "pointers directly"));
    out.print("   bound exc is ");
    out.print(doBoundCheck ? runtimeBound ? "runtime" : "native" : "not checked");
    out.print(", null exc is ");
    out.println(runtimeNull ? "checked" : "native");
    out.print("   maxInlineLevels==");
    out.print(inlineLevels);
    out.print(", maxStmtAutoInline==");
    out.println(maxStmtAutoInline);
    
    //parse everything
    if (timing) {
      out.print("(t1) ");
      t1=osio.getTimeInfo();
    }
    for (i=0; i<argv.length; i++) error|=!fa.scanparse(argv[i]);
    if (compressedImage!=null) {
      error|=!fa.addByteArray(compressedImage.memBlock, 0, compressedImage.memBlockLen, "comprImg");
      error|=!fa.addByteArray(compressedImageInfo, 0, compressedImageInfo.length, "comprInfo");
    }
    fa.scanparseFinished();
    if (error || err) {
      out.println("There were errors");
      return 1;
    }
    
    //check if minimalistic runtime-entvironment exists
    if (timing) {
      out.print("(t2) ");
      t2=osio.getTimeInfo();
    }
    out.println("Checking environment...");
    if (!fa.checkCompEnvironment()) {
      out.println("...minimalistic environment not existing");
      return 2;
    }
    
    //resolve interfaces of units
    if (timing) {
      out.print("(t3) ");
      t3=osio.getTimeInfo();
    }
    out.println("Resolving unit-interfaces...");
    ulist=unitList;
    while (ulist!=null) {
      error|=!ulist.unit.resolveInterface(this);
      ulist=ulist.next;
    }
    if (error || err) {
      out.println("...resolving interfaces failed");
      return 3;
    }
    
    //resolve method-blocks and update imports
    if (timing) {
      out.print("(t4) ");
      t4=osio.getTimeInfo();
    }
    out.println("Resolving methods and updating imports...");
    ulist=unitList;
    while (ulist!=null) {
      error|=!ulist.unit.resolveMethodBlocks(this);
      ulist=ulist.next;
    }
    if (error || err) {
      out.println("...resolving method-blocks or updating imports failed");
      return 4;
    }
    
    //re-check all the offset, assign unassigned ones and updatge references
    if (timing) {
      out.print("(t5) ");
      t5=osio.getTimeInfo();
    }
    out.println("Assigning offsets and preparing descriptors...");
    mem.alignBlock(alignBlockMask);
    //resolve java.lang.Object and java.rte.SClassDesc and update vice-versa
    error|=!langRoot.unitDest.assignOffsets(false, this);
    error|=!rteSClassDesc.assignOffsets(false, this);
    error|=!langRoot.unitDest.assignOffsets(true, this);
    error|=!rteSClassDesc.assignOffsets(true, this);
    error|=!rteSClassDesc.genDescriptor(this); //SClassDesc must be the first allocated object
    error|=!langRoot.unitDest.genDescriptor(this);
    //prepare interface descriptors and maps
    error|=!rteSIntfDesc.assignOffsets(true, this);
    error|=!rteSIntfDesc.genDescriptor(this);
    error|=!rteSIntfMap.assignOffsets(true, this);
    error|=!rteSIntfMap.genDescriptor(this);
    if (!leanRTE) rteSMthdBlock.modifier|=Unit.MA_ACCSSD;
    //do all the others
    ulist=unitList;
    while (ulist!=null) {
      error|=!ulist.unit.assignOffsets(true, this);
      error|=!ulist.unit.genDescriptor(this);
      ulist=ulist.next;
    }
    if (error || err) {
      out.println("...assigning offsets failed");
      return 4;
    }
    
    //get offsets of variables and methods
    if (timing) {
      out.print("(t6) ");
      t6=osio.getTimeInfo();
    }
    out.println("Check environment-structure of languages...");
    if (!fa.checkLangEnvironment()) return 5; //output already done
    if (timing) {
      out.print("(t7) ");
      t7=osio.getTimeInfo();
    }
    
    //build constant objects
    out.println("Building constant objects and strings...");
    mem.alignBlock(alignBlockMask);
    if (embedded) { //create byte array for initialization of class variables
      ramSize=(ramSize+arch.allocClearBits)&~arch.allocClearBits;
      if ((ramInitAddr=mem.allocateArray(ramSize, 1, 1, StdTypes.T_BYTE, 0))<0) return 6;
      mem.enterInitObject(ramInitAddr); //remember address if neccessary
      ramInitAddr+=rteSArray.instScalarTableSize; //point to first element insted of header
    }
    constMemorySize=-mem.getCurrentAllocAmountHint(); //remember current hint
    ulist=unitList;
    while (ulist!=null) {
      error|=!ulist.unit.genConstObj(this);
      ulist=ulist.next;
    }
    constMemorySize+=mem.getCurrentAllocAmountHint(); //now constMemorySize is valid
    if (err) {
      out.println("There were errors");
      return 6;
    }
    
    //create descriptors and generate code
    if (timing) {
      out.print("(t8) ");
      t8=osio.getTimeInfo();
    }
    out.println("Generating descriptors and code...");
    mem.alignBlock(alignBlockMask);
    ulist=unitList;
    while (ulist!=null) {
      uCnt++;
      error|=!ulist.unit.genOutput(this);
      ulist=ulist.next;
    }
    if (error || err || arch.err) {
      out.println("There were errors");
      return 6;
    }
    
    //include symbol information if wanted
    if (timing) {
      out.print("(t9) ");
      t9=osio.getTimeInfo();
    }
    if (symGen!=null) {
      symGenSize=-mem.getCurrentAllocAmountHint(); //remember current hint
      out.println("Generating symbol-information...");
      mem.alignBlock(alignBlockMask);
      if (!symGen.generateSymbols(unitList, this)) { //SymbolInformer may increase symGenSize if wanted
        out.println("There were errors");
        return 7;
      }
      symGenSize+=mem.getCurrentAllocAmountHint(); //now symGenSize is valid
    }
    else if (timing) out.println("(symbol-information are skipped)");
    
    //enter information of method to start in image
    if (timing) {
      out.print("(t10) ");
      t10=osio.getTimeInfo();
    }
    out.println("Entering startup-information...");
    if (!fa.enterStartupInfo()) return 8; //output already done
    
    if (memSize!=0) {
      //write output to disk if there is an output-formatter (ie. there is no compression)
      if (imgOut!=null) {
        out.println("Writing output to target...");
        if (!imgOut.writeOutput(bootMem, compressedImage, compressedImageOrigLen)) {
          out.println("Error writing image");
          return 9;
        }
      }
      else out.println("No output written so far.");
    }
    else out.println("Objects stored in heap.");
    
    //done it, print some statistics and exit with errorcode==0
    out.println("Image finalized.");
    out.print("Created ");
    out.print(mem.objectsAllocated);
    out.print(" objects for ");
    out.print(uCnt);
    out.print(" units, ");
    out.print(mthdCount);
    out.print(" methods and ");
    out.print(stringCount);
    out.println(" strings.");
    out.print("Included ");
    printSize(mthdCodeSize);
    out.print(" code and ");
    out.print(stringChars);
    out.print(" string-chars (");
    printSize(stringMemBytes);
    out.print(")");
    if (memSize!=0) {
      out.print(" in an ");
      printSize(bootMem.memBlockLen);
      out.print(" image");
    }
    out.println(".");
    if (embedded) {
      out.print("Embedded mode with ");
      out.print(ramSize+(embConstRAM ? constMemorySize : 0));
      out.print(" b RAM at 0x");
      out.printHexFix(ramStart, 8);
      out.print(" (first free: ");
      out.printHexFix(ramStart+ramSize+(embConstRAM ? constMemorySize : 0), 8);
      out.println(").");
    }
    if (symGen!=null) {
      out.print("The image contains symbol information which require ");
      printSize(symGenSize);
      out.print(" (+");
      out.print(100*symGenSize/(bootMem.memBlockLen-symGenSize));
      out.println("%).");
    }
    
    arch.printStatistics();
    
    if (timing) {
      out.print("t1-t0="); out.print(t1-t0);
      out.print(", t2-t1="); out.print(t2-t1);
      out.print(", t3-t2="); out.println(t3-t2);
      out.print("t4-t3="); out.print(t4-t3);
      out.print(", t5-t4="); out.print(t5-t4);
      out.print(", t6-t5="); out.println(t6-t5);
      out.print("t7-t6="); out.print(t7-t6);
      out.print(", t8-t7="); out.print(t8-t7);
      out.print(", t9-t8="); out.println(t9-t8);
      out.print("t10-t9="); out.println(t10-t9);
    }
    
    //if compression is enabled, compress image and start second run
    if (compressor!=null) {
      //create environment for decompressor
      decompressor=new Context(osio);
      decompressor.compressedImage=new BootableImage();
      decompressor.compressedImage.memBlock
        =new byte[decompressor.compressedImageOrigLen=bootMem.memBlockLen];
      decompressor.compressedImage.baseAddress=bootMem.baseAddress;
      decompressor.compressedImage.startUnit=bootMem.startUnit;
      decompressor.compressedImage.startCode=bootMem.startCode;
      //get compressed image with information block
      decompressor.compressedArchSpecialValue=arch.getSpecialValue();
      decompressor.compressedImageInfo=bootMem.getInfoBlock(compressorName);
      out.print("Compressing... ");
      if ((i=compressor.compress(bootMem.memBlock, bootMem.memBlockLen,
          decompressor.compressedImage.memBlock, bootMem.memBlockLen))>=bootMem.memBlockLen || i<=0) {
        out.println("could not compress image below original size");
        return 10;
      }
      //output statistics and start compilation of decompressor
      out.print(bootMem.memBlockLen);
      out.print(" to ");
      out.print(i);
      out.print(" bytes (");
      out.print(100*(decompressor.compressedImage.memBlockLen=i)/bootMem.memBlockLen);
      out.println("%) done");
      if ((i=decompressor.compile(secArgv, null))!=0) return i;
      //check overlapping areas
      if (decompressor.bootMem.baseAddress<=bootMem.baseAddress) { //decompressor before us
        if (decompressor.bootMem.baseAddress+decompressor.bootMem.memBlockLen>=bootMem.baseAddress) {
          out.println("warning: lower decompressor end is behind compressed start");
        }
      }
      else { //decompressor behind us
        if (bootMem.baseAddress+bootMem.memBlockLen>=decompressor.bootMem.baseAddress) {
          out.println("warning: higher decompressor start is below compressed end");
        }
      }
    }
    
    //everything done
    return 0;
  }
  
  private void printSize(int size) {
    if (size>=32<<10) {
      out.print((size+1023)>>>10);
      out.print(" kb");
    }
    else {
      out.print(size);
      out.print(" b");
    }
  }
  
  private boolean getParameters(String[] argv) {
    int curPar, i, j;
    DebugWriter dbw;
    
    if (argv==null) return false; //empty parameter not supported
    for (curPar=0; curPar<argv.length; curPar++) if (argv[curPar]!=null
        && argv[curPar].length()>0 && argv[curPar].charAt(0)=='-') {
      if (argv[curPar].length()!=2) {
        out.print("Invalid parameter \"");
        out.print(argv[curPar]);
        out.println("\", available options:");
        printHelp();
        return false;
      }
      switch (argv[curPar].charAt(1)) {
        case 'i':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-i\"");
            return false;
          }
          if (!getParameters(getTokensFromFile(argv[curPar]))) return false;
          break;
        case 'V':
          timing=true;
          break;
        case 'v':
          verbose=true;
          break;
        case 'L':
          if (dynaMem) {
            out.println("Can not combine \"-L\" with \"-m\"");
            return false;
          }
          leanRTE=streamline=true;
          break;
        case 'l':
          if (dynaMem) {
            out.println("Can not combine \"-l\" with \"-m\"");
            return false;
          }
          streamline=true;
          break;
        case 'w':
          alternateObjNew=true;
          break;
        case 'M':
          if (streamline || embedded) {
            out.println("Can not combine \"-M\" with \"-l\" or \"-e\"");
            return false;
          }
          indirScalars=dynaMem=true;
          break;
        case 'm':
          if (streamline || embedded) {
            out.println("Can not combine \"-m\" with \"-l\" or \"-e\"");
            return false;
          }
          dynaMem=true;
          break;
        case 'H':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-H\"");
            return false;
          }
          compiledFiledListname=argv[curPar];
          break;
        case 'h':
          assignHeapCall=true;
          break;
        case 'c':
          assignCall=true;
          break;
        case 'C':
          doArrayStoreCheck=false;
          break;
        case 'B':
          doBoundCheck=false;
          break;
        case 'b':
          runtimeBound=true;
          break;
        case 'x':
          globalStackExtreme=true;
          break;
        case 'n':
          runtimeNull=true;
          break;
        case 'N':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-N\"");
            return false;
          }
          startMethod=argv[curPar];
          break;
        case 'g':
          genAllUnitDesc=true;
          break;
        case 'G':
          genAllMthds=true;
          break;
        case 'd':
          debugCode=true;
          break;
        case 'D':
          if (curPar+2>=argv.length) {
            out.println("Missing parameter for \"-D\"");
            return false;
          }
          if ((dbw=DebugFactory.getDebugWriter(argv[curPar+1], argv[curPar+2], this))==null) {
            out.println("Invalid parameter or unknown debug info writer, currently available:");
            DebugFactory.printKnownDebugWriters(out);
            return false;
          }
          argv[curPar++]=null; //get next parameter, too
          argv[curPar++]=null; //get next parameter, too
          dbw.nextWriter=debugWriter;
          debugWriter=dbw;
          break;
        case 'W':
          printCode=true;
          break;
        case 'r':
          structAsReference=true;
          break;
        case 'R':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-R\"");
            return false;
          }
          if ("nat".equals(argv[curPar])) returnMissingNativeNotRTE=true;
          else if (!"rte".equals(argv[curPar])) { 
            out.println("Invalid parameter for \"-R\", please specify \"rte\" or \"nat\"");
            return false;
          }
          codeReturnMissing=true;
          break;
        case 'F':
          globalProfiling=true;
          break;
        case 'f':
          noThrowFrames=true;
          break;
        case 'Y':
          prefereNativeReal=true;
          break;
        case 'y':
          byteString=true;
          break;
        case 'k':
          noInlineMthdObj=true;
          break;
        case 'K':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-K\"");
            return false;
          }
          if ((alignBlockMask=parseInt(argv[curPar]))<0) {
            out.println("Invalid number in parameter for \"-K\"");
            return false;
          }
          break;
        case 'P':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-P\"");
            return false;
          }
          standardHeader=false;
          if (!argv[curPar].equals("none")) headerFile=argv[curPar];
          break;
        case 'p':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-p\"");
            return false;
          }
          debugPrefix=argv[curPar];
          break;
        case 'I':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-I\"");
            return false;
          }
          if ((inlineLevels=parseInt(argv[curPar]))<0) {
            out.println("Invalid number in parameter for \"-I\"");
            return false;
          }
          break;
        case 'S':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-S\"");
            return false;
          }
          if ((maxStmtAutoInline=parseInt(argv[curPar]))<0) {
            out.println("Invalid number in parameter for \"-S\"");
            return false;
          }
          break;
        case 's':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-s\"");
            return false;
          }
          if ((memSize=parseInt(argv[curPar]))==-1) {
            out.println("Invalid number in parameter for \"-s\"");
            return false;
          }
          break;
        case 'e':
          if (dynaMem) {
            out.println("Can not combine \"-e\" with \"-m\"");
            return false;
          }
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-e\"");
            return false;
          }
          if ((ramStart=parseInt(argv[curPar]))==-1) {
            out.println("Invalid number in parameter for \"-e\"");
            return false;
          }
          embedded=true;
          break;
        case 'E':
          if (!embedded) {
            out.println("Please give \"-e\" before \"-E\"");
            return false;
          }
          embConstRAM=true;
          break;
        case 'a':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-a\"");
            return false;
          }
          if ((memStart=parseInt(argv[curPar]))==-1) {
            out.println("Invalid number in parameter for \"-a\"");
            return false;
          }
          if ((memStart&7)!=0) {
            out.println("Address of image has to be at least 8-byte-aligned");
            return false;
          }
          break;
        case 'A':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-A\"");
            return false;
          }
          if (compressedImage==null) {
            out.println("Parameter \"-A\" is not valid without compressed image");
            return false;
          }
          if (argv[curPar].length()!=1) {
            out.println("Invalid parameter for \"-A\"");
            return false;
          }
          if (memStart==-1) {
            out.println("No \"-a\" parameter given yet, please give before \"-A\"");
            return false;
          }
          j=0;
          for (i=0; i<32; i++) if ((memStart&(1<<i))!=0) j++;
          if (j!=1) {
            out.println("Alignment boundary has to be power of 2");
            return false;
          }
          memStart--; //alignment uses bitmask
          switch (argv[curPar].charAt(0)) {
            case 'r':
              memStart=((compressedImage.baseAddress+compressedImageOrigLen+memStart)&~memStart)
                |(compressedImage.baseAddress&memStart);
              break;
            case 'a':
              memStart=(compressedImage.baseAddress+compressedImageOrigLen+memStart)&~memStart;
              break;
            default:
              out.println("Unknown parameter for \"-A\"");
              return false;
          }
          break;
        case 't':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-t\"");
            return false;
          }
          if ((arch=ArchFactory.getArchitecture(argv[curPar]))==null) {
            out.println("Unknown target architecture, currently available:");
            ArchFactory.printKnownArchitectures(out);
            return false;
          }
          break;
        case 'T':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-T\"");
            return false;
          }
          if (arch==null) {
            out.println("No architecture specified yet, please give before \"-T\"");
            return false;
          }
          if (!arch.setParameter(argv[curPar], out)) return false; 
          break;
        case 'o':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-o\"");
            return false;
          }
          if ((imgOut=OutputFactory.getOutputFormat(argv[curPar]))==null) {
            out.println("Unknown output format, currently available:");
            OutputFactory.printKnownOutputFormats(out);
            return false;
          }
          break;
        case 'O':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-O\"");
            return false;
          }
          if (imgOut==null) {
            out.println("No output format specified yet, please give before \"-O\"");
            return false;
          }
          if (!imgOut.setParameter(argv[curPar], out)) return false; 
          break;
        case 'u':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-u\"");
            return false;
          }
          if ((symGen=SymbolFactory.getSymbolGenerator(argv[curPar]))==null) {
            out.println("Unknown output format, currently available:");
            SymbolFactory.printKnownOutputFormats(out);
            return false;
          }
          break;
        case 'U':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing parameter for \"-U\"");
            return false;
          }
          if (symGen==null) {
            out.println("No output format specified yet, please give before \"-U\"");
            return false;
          }
          if (!symGen.setParameter(argv[curPar], out)) return false; 
          break;
        case 'z':
          argv[curPar++]=null; //get next parameter, too
          if (curPar==argv.length) {
            out.println("Missing algorithm for \"-z\"");
            return false;
          }
          if ((compressor=CompressionFactory.getCompressor(compressorName=argv[curPar]))==null) {
            out.println("Unknown compressor, currently available:");
            CompressionFactory.printKnownCompressors(out);
            return false;
          }
          argv[curPar++]=null; //remove compressor name
          //snip all the following parameters to a new string array for a second compilation
          secArgv=new String[argv.length-curPar];
          for (i=curPar; i<argv.length; i++) {
            secArgv[i-curPar]=argv[i];
            argv[i]=null;
          }
          curPar=argv.length-1; //terminate parameter detection loop
          break;
        default:
          out.print("Unknown parameter \"");
          out.print(argv[curPar].charAt(1));
          out.println("\", available options:");
          printHelp();
          return false;
      }
      argv[curPar]=null;
    }
    return true;
  }
  
  private void printInfo() {
    out.println("SJC by Stefan Frenz, released under GPL");
    out.println("The software comes \"as is\" and without any warranty");
  }
  
  private void printHelp() {
    printInfo();
    //available chars:    abcdefghijklmnopqrstuvwxyz
    //unused lower case: |         j      q         |
    //unused upper case: |         J      Q      X Z|
    out.println("Please give at least one file to compile");
    out.println("Available options are:");
    out.println(" -i FILE    get options from FILE");
    out.println(" -V         verbose timing information");
    out.println(" -v         verbose mode, print more infos");
    out.println(" -L         flat rte (no obj.), implies \"-l\"");
    out.println(" -l         streamline objects, inhibits \"-m\"");
    out.println(" -w         use alternate Object and newInstance");
    out.println(" -M         build indirect scalars, implies \"-m\"");
    out.println(" -m         generate movable code and descriptors");
    out.println(" -H FILE    output file-listing of compiled files");
    out.println(" -h         generate call for heap pointer assign");
    out.println(" -c         generate call for every pointer assign");
    out.println(" -C         completely skip all array store checks");
    out.println(" -B         completely skip all array bound checks");
    out.println(" -b         generate runtime call on bound exception");
    out.println(" -x         generate stack extreme check for methods");
    out.println(" -n         generate runtime call on nullpointer use");
    out.println(" -N PK.U.MT use method MT in unit PK.U as entry method");
    out.println(" -g         generate all unit-descriptors (don't skip)");
    out.println(" -G         generate all mthd-code-bodies (don't skip)");
    out.println(" -d         create binary debug output for each method");
    out.println(" -D DW FILE debug writer DW with FILE (sym is default)");
    out.println(" -W         print all method's instructions if possible");
    out.println(" -r         treat structs as objects, store at references");
    out.println(" -R WAY     insert return missing call in WAY=nat or =rte");
    out.println(" -F         insert calls to runtime-profiler at each method");
    out.println(" -f         do not generate code for throw-frames (no catch)");
    out.println(" -Y         prefer native float/double usage inside compiler");
    out.println(" -y         use byte-values instead of char-values in String");
    out.println(" -k         skip method code objects for inline-marked methods");
    out.println(" -K MASK    align blocks of allocated objects with mask MASK");
    out.println(" -P FILE    special (or \"none\") instead of standard header");
    out.println(" -p PREF    naming prefix for all debug files to be written");
    out.println(" -I LEVL    specify maximum level of method inlining (def==3)");
    out.println(" -S AMNT    maximum statement amount for auto-inline (def==0)");
    out.println(" -s SIZE    specify maximum size of memory image to SIZE");
    out.println(" -s 0       do not use image, allocate in bootstrap mode");
    out.println(" -e ADDR    embedded mode, set static vars' RAM address");
    out.println(" -E         constant objects through RAM, needs \"-e\"");
    out.println(" -a ADDR    specify start address of memory image");
    out.println(" -t ARCH    specify target architecture");
    out.println(" -T APRM    parameter for architecture");
    out.println(" -o OFMT    specify output format");
    out.println(" -O OPRM    parameter for output");
    out.println(" -u UFMT    specify symbol generator");
    out.println(" -U UPRM    parameter for symbol generator");
    out.println(" -z CALG X  compression, following params X are for decompressor");
    out.println("    -A        change mode of given address to one of these:");
    out.println("       r      relative: align to boundary, keep lower bits");
    out.println("       a      align: align to next boundary");
    out.println("Possible input types are:");
    FrontAdmin.printKnownLanguages(out);
    out.println("Possible architectures are:");
    ArchFactory.printKnownArchitectures(out);
    out.println("Possible output formats are:");
    OutputFactory.printKnownOutputFormats(out);
    out.println("Possible symbol generators are:");
    SymbolFactory.printKnownOutputFormats(out);
    out.println("Possible compression modules are:");
    CompressionFactory.printKnownCompressors(out);
    out.println("Possible debug file writers are:");
    DebugFactory.printKnownDebugWriters(out);
  }
  
  private String[] getTokensFromFile(String filename) {
    TextReader r=osio.getNewTextReader();
    boolean inString=false, nextESC=false;
    int bufLen=0, tmpLen, tokCnt=0;
    char[] buffer=new char[30], tmpBuf;
    char c;
    StringList tokens=null, lastTok=null;
    String[] ret;
    
    if (!r.initData(filename)) {
      out.print("error opening file ");
      out.println(filename);
      return null;
    }
    while (true) { //read until end of file or end of line
      if ((c=r.nextChar)=='\n' || c=='\r' || c=='\0') { //finished reading
        if (bufLen>0) { //build remaining string
          if (tokens==null) lastTok=tokens=new StringList(new String(buffer, 0, bufLen));
          else lastTok=new StringList(lastTok, new String(buffer, 0, bufLen));
          tokCnt++;
        }
        break; //stop loop
      }
      if (nextESC) { buffer[bufLen++]=c; nextESC=false; } //after escape character
      else if (c=='\\') nextESC=true; //escape character
      else if (inString && c=='"') inString=false; //end of string
      else if (c=='"') inString=true; //start of string
      else if (inString || c!=' ') buffer[bufLen++]=c; //normal character
      else if (bufLen>0) { //option finished, build string
        if (tokens==null) lastTok=tokens=new StringList(new String(buffer, 0, bufLen));
        else lastTok=new StringList(lastTok, new String(buffer, 0, bufLen));
        bufLen=0;
        tokCnt++;
      }
      r.readChar(); //get next character
      if (bufLen+1>=buffer.length) { //buffer exceeded, build new one
        tmpBuf=new char[buffer.length];
        for (tmpLen=0; tmpLen<bufLen; tmpLen++)
          tmpBuf[tmpLen]=buffer[tmpLen];
        buffer=tmpBuf;
      }
    }
    if (inString || nextESC) {
      out.print("unfinished string or escape sequence in file ");
      out.println(filename);
      return null;
    }
    //everything ok, got all tokens, now build array of tokens
    ret=new String[tokCnt];
    for (tmpLen=0; tmpLen<tokCnt; tmpLen++) {
      ret[tmpLen]=tokens.str;
      tokens=tokens.next;
    }
    return ret;
  }
  
  public void addUnit(Unit u) {
    UnitList newOne;
    
    newOne=new UnitList(u);
    if (lastUnit!=null) {
      lastUnit.next=newOne;
      lastUnit=newOne;
    }
    else unitList=lastUnit=newOne;
  }
  
  public void writeSymInfo() {
    DebugWriter dbw;
    
    if (verbose) out.println("Writing debug info...");
    if (debugWriter==null) debugWriter=DebugFactory.getDebugWriter(
        null, debugPrefix.concat("syminfo.txt"), this); //get standard debug writer
    dbw=debugWriter;
    while (dbw!=null) {
      writeSymInfo(dbw);
      dbw.finalizeImageInfo();
      dbw=dbw.nextWriter;
    }
    if (verbose) out.println("Debug info written.");
  }
  
  private void writeSymInfo(DebugWriter symOut) {
    UnitList ulist;
    
    symOut.startImageInfo(compressedImage!=null);
    
    symOut.globalMemoryInfo(bootMem.baseAddress, bootMem.memBlockLen);
    symOut.globalMethodInfo(mthdCodeSize, mthdCount);
    symOut.globalStringInfo(stringCount, stringChars, stringMemBytes);
    symOut.globalRAMInfo(ramInitAddr, ramSize, constMemorySize);
    symOut.globalSymbolInfo(symGenSize);
    
    ulist=unitList;
    while (ulist!=null) {
      ulist.unit.writeDebug(this, symOut);
      ulist=ulist.next;
    }
    
    if (decompressor!=null) decompressor.writeSymInfo(symOut);
  }
  
  public void addFile(StringList fileName) { //recycle list, next is not used
    if (fileList!=null) fileName.tablePos=fileList.tablePos+1;
    else fileName.tablePos=1;
    fileName.next=fileList;
    fileList=fileName;
  }
  
  public String getNameOfFile(int fileID) {
    StringList s;
    int first, last;
    
    if (fileID!=-1) {
      s=fileList;
      while (s!=null) {
        if (s.tablePos==fileID) {
          first=s.str.lastIndexOf('/');
          if (first==-1) first=s.str.lastIndexOf('\\');
          first++;
          last=s.str.lastIndexOf('.');
          return s.str.substring(first, last); 
        }
        s=s.next;
      }
    }
    return null;
  }
  
  public void printPos(int fileID, int line, int col) {
    StringList s;
    
    if (fileID==-1) {
      out.print("internal init");
      if (line!=-1 || col!=-1) {
        out.print(':');
        out.print(line);
        out.print(':');
        out.print(col);
      }
    }
    else {
      s=fileList;
      while (s!=null) {
        if (s.tablePos==fileID) {
          out.print(s.str);
          if (line!=-1 || col!=-1) {
            out.print(':');
            out.print(line);
            out.print(':');
            out.print(col);
          }
          return;
        }
        s=s.next;
      }
      out.print("#unkown file#");
    }
  }
  
  public int allocateString(String what) {
    int strAddr, cArAddr, off, len, tmp, res=0;
    
    if (what==null || err) return 0;
    len=what.length();
    if (langString.inlArr!=null) { //inline array
      //allocate space
      tmp=len;
      if (!byteString) tmp=tmp<<1;
      strAddr=mem.allocate(
          langString.instScalarTableSize+tmp, 0,
          langString.instRelocTableEntries, langString.outputAddr);
      if (strAddr<0) {
        out.println("error allocating memory while creating inline strings");
        err=true;
        return 0;
      }
      //enter pointers and infos
      arch.putInt(strAddr+langStrCnt, len);
      if (embConstRAM) res=strAddr-ramInitAddr+ramStart;
      else res=strAddr;
      //enter characters
      cArAddr=strAddr+langString.instScalarTableSize;
    }
    else { //normal array
      //allocate space
      strAddr=mem.allocate(
        langString.instScalarTableSize, langString.instIndirScalarTableSize,
        langString.instRelocTableEntries, langString.outputAddr);
      if (strAddr<0) {
        out.println("error allocating memory while creating strings");
        err=true;
        return 0;
      }
      if (byteString) cArAddr=mem.allocateArray(len, 1, 1, StdTypes.T_BYTE, 0);
      else cArAddr=mem.allocateArray(len, 1, 2, StdTypes.T_CHAR, 0);
      if (cArAddr<0) {
        out.println("error allocating memory while creating array for string");
        err=true;
        return 0;
      }
      //enter pointers and infos
      if (embConstRAM) arch.putRef(strAddr-langStrVal*arch.relocBytes,
          cArAddr-ramInitAddr+ramStart);
      else arch.putRef(strAddr-langStrVal*arch.relocBytes, cArAddr);
      if (langStrCnt>=0) arch.putInt(strAddr+langStrCnt, len);
      if (dynaMem) res=strAddr;
      else if (embConstRAM) res=strAddr-ramInitAddr+ramStart;
      else res=strAddr;
      if (indirScalars) cArAddr=mem.getIndirAddrOfLastObject()+rteSArray.instIndirScalarTableSize;
      else cArAddr+=rteSArray.instScalarTableSize;
    }
    //enter characters
    if (byteString) for (off=0; off<len; off++) mem.putByte(cArAddr+off, (byte)what.charAt(off));
    else for (off=0; off<len; off++) arch.putShort(cArAddr+(off<<1), (short)what.charAt(off));
    return res;
  }
  
  public int allocateMultiArray(int curDim, FilledParam sizes,
      int entrySize, int stdType, int extTypeAddr) {
    int i, max, obj, cArAddr;
    
    if (sizes==null || sizes.expr==null) return 0;
    max=sizes.expr.getConstIntValue(this);
    if ((obj=mem.allocateArray(max, curDim, entrySize,
        stdType, extTypeAddr))>=0 && curDim>1) {
      cArAddr=obj-(rteSArray.instRelocTableEntries+1)*arch.relocBytes;
      for (i=0; i<max; i++, cArAddr-=arch.relocBytes)
        arch.putRef(cArAddr, allocateMultiArray(curDim-1, sizes.nextParam,
            entrySize, stdType, extTypeAddr));
    }
    return obj;
  }
  
  private static int parseInt(String s) {
    int ret=0, c, v, sign, pos=0;
    int base=10;
    
    if (s==null || s.length()==0) return -1;
    
    while (s.charAt(pos)==' ' || s.charAt(pos)=='\t') { //skip leading blanks
      if (++pos>=s.length()) return -1;
    }
    
    if (s.charAt(pos)=='-') { //get sign
      sign=-1; pos++;
    }
    else { //default is positive value
      if (s.charAt(pos)=='+') pos++; //accept sign
      sign=1;
    }
    
    //check for hex number
    if (pos+2<s.length() && s.charAt(pos)=='0' && (s.charAt(pos+1)=='x' || s.charAt(pos+1)=='X')) {
      pos+=2;
      base=16;
    }
    
    while (pos<s.length()) {
      c=(int)s.charAt(pos)&0xFF;
      if (c>=48 && c<=57) v=c-48; //0..9
      else if (c>=65 && c<=70) v=c-55; //10..15
      else if (c>=97 && c<=102) v=c-87; //10..15
      else if (c==75 || c==107) { //K/k
        if (pos!=s.length()-1) return -1;
        ret*=1024;
        break;
      }
      else if (c==77 || c==109) { //M/m
        if (pos!=s.length()-1) return -1;
        ret*=1024*1024;
        break;
      }
      else return -1;
      if (v>=base) return -1;
      ret=ret*base+v;
      pos++;
    }
    return sign*ret;
  }
  
  public VrblList getUnusedVrblList(Vrbl vrbl, VrblList next) {
    VrblList result;
    if (emptyVrblLists==null) result=new VrblList(vrbl); //get new VrblList
    else { //recycle old VrblList
      result=emptyVrblLists;
      emptyVrblLists=emptyVrblLists.next;
      result.vrbl=vrbl;
    }
    result.next=next;
    return result;
  }
  
  public VrblList getUnwrittenVars(VrblList toCheck) {
    VrblList result=null;
    while (toCheck!=null) {
      if ((toCheck.vrbl.modifier&AccVar.MF_ISWRITTEN)==0) {
        result=getUnusedVrblList(toCheck.vrbl, result);
      }
      toCheck=toCheck.next;
    }
    return result;
  }
  
  public VrblList getWrittenVarsSortedAndSet(VrblList varList, int replacementFlag) {
    VrblList first=null, last=null, tmp;
    while (varList!=null) {
      if ((varList.vrbl.modifier&AccVar.MF_ISWRITTEN)!=0) {
        tmp=getUnusedVrblList(varList.vrbl, null);
        if (first==null) first=last=tmp;
        else {
          last.next=tmp;
          last=tmp;
        }
        varList.vrbl.modifier=(varList.vrbl.modifier&~AccVar.MF_ISWRITTEN)|replacementFlag;
      }
      varList=varList.next;
    }
    return first;
  }
  
  public void recycleFullVrblList(VrblList toRecycle) {
    while (toRecycle!=null) toRecycle=recycleSingleVrblList(toRecycle);
  }
  
  public VrblList recycleSingleVrblList(VrblList toRecycle) {
    VrblList result;
    result=toRecycle.next;
    toRecycle.next=emptyVrblLists;
    emptyVrblLists=toRecycle;
    return result;
  }

  public void setVarsWrittenState(VrblList varList, int stateFlag) {
    while (varList!=null) {
      varList.vrbl.modifier|=stateFlag;
      varList=varList.next;
    }
  }

  public void setCombinedVarsWrittenState(VrblList masterList, VrblList written1, VrblList written2) {
    boolean flag1, flag2;
    while (masterList!=null) {
      flag1=written1!=null && masterList.vrbl==written1.vrbl;
      flag2=written2!=null && masterList.vrbl==written2.vrbl;
      if (flag1 && flag2) { //variable is written in both true- and false-block
        masterList.vrbl.modifier|=AccVar.MF_ISWRITTEN;
        written1=written1.next;
        written2=written2.next;
      }
      else if (flag1) {
        masterList.vrbl.modifier|=AccVar.MF_MAYBEWRITTEN;
        written1=written1.next;
      }
      else if (flag2) {
        masterList.vrbl.modifier|=AccVar.MF_MAYBEWRITTEN;
        written2=written2.next;
      }
      //else: variable is written neither in written1 nor in written2
      masterList=masterList.next;
    }
  }
}
