/* Copyright (C) 2007, 2008, 2009 Stefan Frenz and Patrick Schmidt
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package symbols;

import osio.TextPrinter;
import compbase.*;

/**
 * BootStrap: generate output for all compiler relevant symbol information
 * 
 * @author S. Frenz, P. Schmidt
 * @version 090718 adopted move of modifier flags from Vrbl to AccVar
 *  version 090609 fixed putRef instead of putInt in genVrbl
 *  version 090207 added copyright notice
 *  version 080614 adopted changed Unit.searchVariable
 *  version 080327 fixed filling of method symHint
 *  version 080115 adopted changes in SymbolInformer.generateSymbols, using SymbolInformer.checkVrbl
 *  version 080114 removed checkParameter
 *  version 070920 stricter test of root variable declaration, fixed root variable setting in dynaMem mode
 *  version 070816 removed osio from compiler-exlusion list, added parameter to avoid null-unitDest
 *  version 070801 added Unit.offset*
 *  version 070731 adopted QualIDList, adopted renaming of id to name
 *  version 070730 fixed option "noCompilerPacks" to include packages but to exclude units
 *  version 070729 added option "noCompilerPacks" to reduce size of added symbol information
 *  version 070729 removed no longer needed prepareUnits, added type check for root-variable, fixed some bugs
 *  version 070727 initial version
 */

public class BootStrapSymbols extends SymbolInformer {
  private boolean noCompilerPacks;
  private Unit packUnit, unitUnit, mthdUnit, vrblUnit, paramUnit, typeRefUnit, qualIDUnit, qualIDListUnit;
  private Unit stringListUnit;
  private int packNamePos, packOuterPos, packSubPacksPos, packNextPackPos, packUnitsPos;
  private int unitNamePos, unitExtsIDPos, unitExtsImplIDListPos, unitModifierPos;
  private int unitNextUnitPos, unitPackPos, unitMthdsPos, unitVarsPos, unitInlArrPos, unitOutputAddrPos;
  private int unitExplConstrPos, unitExplStdConstrPos, unitInitStatPos, unitInitDynaPos, unitStatRelocTableEntriesPos;
  private int unitStatScalarTableSizePos, unitClssRelocTableEntriesPos, unitClssScalarTableSizePos;
  private int unitInstRelocTableEntriesPos, unitInstScalarTableSizePos, unitInstIndirScalarTableSizePos;
  private int unitOffsetsAssignedPos, unitOutputGeneratedPos;
  private int mthdRetTypePos, mthdOwnerPos, mthdMarkerPos, mthdJumpTableIndexPos, mthdIsConstructorPos;
  private int mthdOvldMthdPos, mthdRedirectPos, mthdOutputAddrPos, mthdCodeOutputAddrPos, mthdParamPos;
  private int mthdNamePos, mthdModifierPos, mthdNextMthdPos;
  private int vrblOwnerPos, vrblModifierPos, vrblNextVrblPos, vrblNamePos, vrblTypePos;
  private int vrblMinSizePos, vrblVarPosPos, vrblLocationPos;
  private int paramNextParamPos, paramNamePos, paramTypePos, paramMinSizePos, paramVarPosPos;
  private int paramLocationPos;
  private int typeRefBaseTypePos, typeRefArrDimPos, typeRefInlinedArraySizePos, typeRefQidPos;
  private int qualIDNamePos, qualIDTypePos, qualIDPackDestPos, qualIDUnitDestPos;
  private int qualIDListQidPos, qualIDListNextPos;
  private int stringListStrPos, stringListNextPos, stringListTablePosPos;
  private TypeRefPool typeRefPool;
  
  public static void printValidParameters(TextPrinter v) {
    v.println("   noCompilerPacks   exclude packages of compiler from symbol information");
  }
  
  public boolean setParameter(String parm, TextPrinter v) {
    if (parm.equals("noCompilerPacks")) {
      noCompilerPacks=true;
      return true;
    }
    v.println("BootStrapSymbols accepts only the following parameters:");
    printValidParameters(v);
    return false;
  }

  public boolean generateSymbols(UnitList newUnits, Context ictx) {
    int rootAddr;
    Unit myUnit;
    Vrbl rootVar;
    StringList list;
    Pack compbasePack;
    
    ctx=ictx;
    typeRefPool=new TypeRefPool();
    //check basics
    list=new StringList("symbols");
    list.next=new StringList("BootStrap");
    if ((myUnit=ctx.root.searchUnit(list))==null
        || (myUnit.modifier&(Unit.M_ABSTR|Unit.M_INDIR|Unit.M_STRUCT))!=0) {
      ctx.out.println("unit symbols.BootStrap required for BootStrapSymbols (non-abstract, non-indir, non-struct)");
      return false;
    }
    if ((compbasePack=ctx.root.searchSubPackage("compbase"))==null
        || (packUnit=compbasePack.searchUnit("Pack"))==null
        || (unitUnit=compbasePack.searchUnit("UnitDummy"))==null
        || (mthdUnit=compbasePack.searchUnit("MthdDummy"))==null
        || (vrblUnit=compbasePack.searchUnit("Vrbl"))==null
        || (paramUnit=compbasePack.searchUnit("Param"))==null
        || (typeRefUnit=compbasePack.searchUnit("TypeRef"))==null
        || (qualIDUnit=compbasePack.searchUnit("QualID"))==null
        || (qualIDListUnit=compbasePack.searchUnit("QualIDList"))==null
        || (stringListUnit=compbasePack.searchUnit("StringList"))==null) {
      ctx.out.println("required packages or units not found");
      return false;
    }
    if ((rootVar=myUnit.searchVariable("root", ctx))==null
        || (rootVar.modifier&(AccVar.M_FIN|AccVar.M_STAT|AccVar.M_STRUCT))!=AccVar.M_STAT
        || rootVar.type.baseType!=TypeRef.T_QID || rootVar.type.qid.unitDest!=packUnit
        || rootVar.type.arrDim!=0) {
      ctx.out.println("variable \"static Pack root\" needed in symbols.BootStrap");
      return false;
    }
    if ((packNamePos=checkVrbl(packUnit, "name"))==-1
        || (packOuterPos=checkVrbl(packUnit, "outer"))==-1
        || (packSubPacksPos=checkVrbl(packUnit, "subPacks"))==-1
        || (packNextPackPos=checkVrbl(packUnit, "nextPack"))==-1
        || (packUnitsPos=checkVrbl(packUnit, "units"))==-1
        || (unitNamePos=checkVrbl(unitUnit, "name"))==-1
        || (unitExtsIDPos=checkVrbl(unitUnit, "extsID"))==-1
        || (unitExtsImplIDListPos=checkVrbl(unitUnit, "extsImplIDList"))==-1
        || (unitModifierPos=checkVrbl(unitUnit, "modifier"))==-1
        || (unitNextUnitPos=checkVrbl(unitUnit, "nextUnit"))==-1
        || (unitPackPos=checkVrbl(unitUnit, "pack"))==-1
        || (unitMthdsPos=checkVrbl(unitUnit, "mthds"))==-1
        || (unitVarsPos=checkVrbl(unitUnit, "vars"))==-1
        || (unitInlArrPos=checkVrbl(unitUnit, "inlArr"))==-1
        || (unitOutputAddrPos=checkVrbl(unitUnit, "outputAddr"))==-1
        || (unitExplConstrPos=checkVrbl(unitUnit, "explConstr"))==-1
        || (unitExplStdConstrPos=checkVrbl(unitUnit, "explStdConstr"))==-1
        || (unitInitStatPos=checkVrbl(unitUnit, "initStat"))==-1
        || (unitInitDynaPos=checkVrbl(unitUnit, "initDyna"))==-1
        || (unitStatRelocTableEntriesPos=checkVrbl(unitUnit, "statRelocTableEntries"))==-1
        || (unitStatScalarTableSizePos=checkVrbl(unitUnit, "statScalarTableSize"))==-1
        || (unitClssRelocTableEntriesPos=checkVrbl(unitUnit, "clssRelocTableEntries"))==-1
        || (unitClssScalarTableSizePos=checkVrbl(unitUnit, "clssScalarTableSize"))==-1
        || (unitInstRelocTableEntriesPos=checkVrbl(unitUnit, "instRelocTableEntries"))==-1
        || (unitInstScalarTableSizePos=checkVrbl(unitUnit, "instScalarTableSize"))==-1
        || (unitInstIndirScalarTableSizePos=checkVrbl(unitUnit, "instIndirScalarTableSize"))==-1
        || (unitOffsetsAssignedPos=checkVrbl(unitUnit, "offsetsAssigned"))==-1
        || (unitOutputGeneratedPos=checkVrbl(unitUnit, "outputGenerated"))==-1
        || (mthdNamePos=checkVrbl(mthdUnit, "name"))==-1
        || (mthdRetTypePos=checkVrbl(mthdUnit, "retType"))==-1
        || (mthdNextMthdPos=checkVrbl(mthdUnit, "nextMthd"))==-1
        || (mthdOwnerPos=checkVrbl(mthdUnit, "owner"))==-1
        || (mthdModifierPos=checkVrbl(mthdUnit, "modifier"))==-1
        || (mthdMarkerPos=checkVrbl(mthdUnit, "marker"))==-1
        || (mthdJumpTableIndexPos=checkVrbl(mthdUnit, "jumpTableIndex"))==-1
        || (mthdIsConstructorPos=checkVrbl(mthdUnit, "isConstructor"))==-1
        || (mthdOvldMthdPos=checkVrbl(mthdUnit, "ovldMthd"))==-1
        || (mthdRedirectPos=checkVrbl(mthdUnit, "redirect"))==-1
        || (mthdOutputAddrPos=checkVrbl(mthdUnit, "outputAddr"))==-1
        || (mthdCodeOutputAddrPos=checkVrbl(mthdUnit, "codeOutputAddr"))==-1
        || (mthdParamPos=checkVrbl(mthdUnit, "param"))==-1
        || (vrblOwnerPos=checkVrbl(vrblUnit, "owner"))==-1 
        || (vrblModifierPos=checkVrbl(vrblUnit, "modifier"))==-1 
        || (vrblNextVrblPos=checkVrbl(vrblUnit, "nextVrbl"))==-1 
        || (vrblNamePos=checkVrbl(vrblUnit, "name"))==-1 
        || (vrblTypePos=checkVrbl(vrblUnit, "type"))==-1 
        || (vrblMinSizePos=checkVrbl(vrblUnit, "minSize"))==-1 
        || (vrblVarPosPos=checkVrbl(vrblUnit, "varPos"))==-1 
        || (vrblLocationPos=checkVrbl(vrblUnit, "location"))==-1
        || (paramNextParamPos=checkVrbl(paramUnit, "nextParam"))==-1
        || (paramNamePos=checkVrbl(paramUnit, "name"))==-1
        || (paramTypePos=checkVrbl(paramUnit, "type"))==-1
        || (paramMinSizePos=checkVrbl(paramUnit, "minSize"))==-1
        || (paramVarPosPos=checkVrbl(paramUnit, "varPos"))==-1
        || (paramLocationPos=checkVrbl(paramUnit, "location"))==-1
        || (typeRefBaseTypePos=checkVrbl(typeRefUnit, "baseType"))==-1
        || (typeRefArrDimPos=checkVrbl(typeRefUnit, "arrDim"))==-1
        || (typeRefInlinedArraySizePos=checkVrbl(typeRefUnit, "inlinedArraySize"))==-1
        || (typeRefQidPos=checkVrbl(typeRefUnit, "qid"))==-1
        || (qualIDNamePos=checkVrbl(qualIDUnit, "name"))==-1
        || (qualIDTypePos=checkVrbl(qualIDUnit, "type"))==-1
        || (qualIDPackDestPos=checkVrbl(qualIDUnit, "packDest"))==-1
        || (qualIDUnitDestPos=checkVrbl(qualIDUnit, "unitDest"))==-1
        || (qualIDListQidPos=checkVrbl(qualIDListUnit, "qid"))==-1
        || (qualIDListNextPos=checkVrbl(qualIDListUnit, "nextQualID"))==-1
        || (stringListStrPos=checkVrbl(stringListUnit, "str"))==-1
        || (stringListNextPos=checkVrbl(stringListUnit, "next"))==-1
        || (stringListTablePosPos=checkVrbl(stringListUnit, "tablePos"))==-1){
      return false;
    }
    //create objects for packages, units, methods and variables; enter root in static var
    if ((rootAddr=genPackage(ctx.root))==0) return false;
    ctx.arch.putRef(ctx.dynaMem ? myUnit.outputAddr-rootVar.varPos*ctx.arch.relocBytes : rootVar.varPos, rootAddr);
    //everything done
    return !ctx.err;
  }
  
  private int alloc(Unit unit) {
    int addr;
    if (ctx.err) return -1; //no error printing if an error occurred already 
    if ((addr=ctx.mem.allocate(unit.instScalarTableSize, unit.instIndirScalarTableSize,
        unit.instRelocTableEntries, unit.outputAddr))<0) {
      ctx.out.print("error allocating ");
      ctx.out.print(unit.name);
      ctx.out.println("-object for BootStrapSymbols");
      ctx.err=true;
    }
    return addr;
  }
  
  private boolean isCompilerPack(Pack pack) {
    String n;
    
    return pack!=null && (n=pack.name)!=null && (n.equals("backend")
        || n.equals("compbase")  || n.equals("compression")
        || n.equals("emulation") || n.equals("frontend")
        || n.equals("memory") || n.equals("output") || n.equals("symbols"));
  }
  
  private int genPackage(Pack pack) {
    int packAddr=0;
    
    if (pack!=null) {
      if (pack.symHint!=0) return pack.symHint;
      if ((packAddr=alloc(packUnit))<0) return 0;
      pack.symHint=packAddr;
      ctx.arch.putRef(packAddr-packNamePos, ctx.allocateString(pack.name));
      ctx.arch.putRef(packAddr-packOuterPos, genPackage(pack.outer));
      if (!noCompilerPacks || !isCompilerPack(pack)) {
        ctx.arch.putRef(packAddr-packUnitsPos, genUnit(pack.units, false));
        ctx.arch.putRef(packAddr-packSubPacksPos, genPackage(pack.subPacks));
      }
      ctx.arch.putRef(packAddr-packNextPackPos, genPackage(pack.nextPack));
    }
    return packAddr;
  }
  
  private int genUnit(Unit unit, boolean avoidNull) {
    int unitAddr=0;
    
    if (unit!=null) {
      if (noCompilerPacks && unit.pack!=null && isCompilerPack(unit.pack.packDest))
        return avoidNull ? genUnit(ctx.langRoot.unitDest, false) : 0;
      if (unit.symHint!=0) return unit.symHint;
      if ((unitAddr=alloc(unitUnit))<0) return 0;
      unit.symHint=unitAddr;
      ctx.arch.putRef(unitAddr-unitNamePos, ctx.allocateString(unit.name));
      ctx.arch.putRef(unitAddr-unitPackPos, genQualID(unit.pack));
      ctx.arch.putRef(unitAddr-unitMthdsPos, genMthd(unit.mthds));
      ctx.arch.putInt(unitAddr+unitModifierPos, unit.modifier);
      ctx.arch.putRef(unitAddr-unitExtsIDPos, genQualID(unit.extsID));
      ctx.arch.putRef(unitAddr-unitExtsImplIDListPos, genQualIDList(unit.extsImplIDList));
      ctx.arch.putInt(unitAddr+unitOutputAddrPos, unit.outputAddr);
      ctx.mem.putByte(unitAddr+unitExplConstrPos, unit.explConstr ? (byte)1 : (byte)0);
      ctx.mem.putByte(unitAddr+unitExplStdConstrPos, unit.explStdConstr ? (byte)1 : (byte)0);
      ctx.arch.putRef(unitAddr-unitVarsPos, genVrbl(unit.vars));
      ctx.arch.putRef(unitAddr-unitInlArrPos, genVrbl(unit.inlArr));
      ctx.arch.putRef(unitAddr-unitInitStatPos, genMthd(unit.initStat));
      ctx.arch.putRef(unitAddr-unitInitDynaPos, genMthd(unit.initDyna));
      ctx.arch.putRef(unitAddr-unitNextUnitPos, genUnit(unit.nextUnit, avoidNull));
      ctx.arch.putInt(unitAddr+unitStatRelocTableEntriesPos, unit.statRelocTableEntries);
      ctx.arch.putInt(unitAddr+unitStatScalarTableSizePos, unit.statScalarTableSize);
      ctx.arch.putInt(unitAddr+unitClssRelocTableEntriesPos, unit.clssRelocTableEntries);
      ctx.arch.putInt(unitAddr+unitClssScalarTableSizePos, unit.clssScalarTableSize);
      ctx.arch.putInt(unitAddr+unitInstRelocTableEntriesPos, unit.instRelocTableEntries);
      ctx.arch.putInt(unitAddr+unitInstScalarTableSizePos, unit.instScalarTableSize);
      ctx.arch.putInt(unitAddr+unitInstIndirScalarTableSizePos, unit.instIndirScalarTableSize);
      ctx.mem.putByte(unitAddr+unitOffsetsAssignedPos, unit.offsetsAssigned ? (byte)1 : (byte)0);
      ctx.mem.putByte(unitAddr+unitOutputGeneratedPos, unit.outputGenerated ? (byte)1 : (byte)0);
    }
    return unitAddr;
  }
  
  private int genMthd(Mthd mthd) {
    int mthdAddr=0;
    
    if (mthd!=null) {
      if (mthd.symHint!=0) return mthd.symHint;
      if ((mthdAddr=alloc(mthdUnit))<0) return 0;
      mthd.symHint=mthdAddr;
      ctx.arch.putRef(mthdAddr-mthdNextMthdPos, genMthd(mthd.nextMthd));
      ctx.arch.putRef(mthdAddr-mthdNamePos, ctx.allocateString(mthd.name));
      ctx.arch.putRef(mthdAddr-mthdOvldMthdPos, genMthd(mthd.ovldMthd));
      ctx.arch.putInt(mthdAddr+mthdModifierPos, mthd.modifier);
      ctx.arch.putRef(mthdAddr-mthdRetTypePos, genTypeRef(mthd.retType));
      ctx.arch.putRef(mthdAddr-mthdOwnerPos, genUnit(mthd.owner, true));
      ctx.arch.putInt(mthdAddr+mthdMarkerPos, mthd.marker);
      ctx.arch.putInt(mthdAddr+mthdJumpTableIndexPos, mthd.jumpTableIndex);
      ctx.mem.putByte(mthdAddr+mthdIsConstructorPos, mthd.isConstructor ? (byte)1 : (byte)0);
      ctx.arch.putRef(mthdAddr-mthdRedirectPos, genMthd(mthd.redirect));
      ctx.arch.putInt(mthdAddr+mthdOutputAddrPos, mthd.outputAddr);
      ctx.arch.putInt(mthdAddr+mthdCodeOutputAddrPos, mthd.codeOutputAddr);
      ctx.arch.putRef(mthdAddr-mthdParamPos, genParam(mthd.param));
    }
    return mthdAddr;
  }
  
  private int genTypeRef(TypeRef typeRef) {
    int typeRefAddr=0;
    if (typeRef!=null) {
      if (typeRef.symHint!=0) return typeRef.symHint;
      if ((typeRefAddr=typeRefPool.findMatchingTypeRef(typeRef))!=0) return typeRefAddr;
      if ((typeRefAddr=alloc(typeRefUnit))<0) return 0;
      typeRef.symHint=typeRefAddr;
      ctx.arch.putInt(typeRefAddr+typeRefBaseTypePos, typeRef.baseType);
      ctx.arch.putInt(typeRefAddr+typeRefArrDimPos, typeRef.arrDim);
      ctx.arch.putInt(typeRefAddr+typeRefInlinedArraySizePos, typeRef.inlinedArraySize);
      ctx.arch.putRef(typeRefAddr-typeRefQidPos, genQualID(typeRef.qid));
    }
    return typeRefAddr;
  }
  
  private int genParam(Param param) {
    int paramAddr=0;
    if (param!=null) {
      if ((paramAddr=alloc(paramUnit))<0) return 0;
      ctx.arch.putRef(paramAddr-paramNextParamPos, genParam(param.nextParam));
      ctx.arch.putRef(paramAddr-paramNamePos, ctx.allocateString(param.name));
      ctx.arch.putRef(paramAddr-paramTypePos, genTypeRef(param.type));
      ctx.arch.putInt(paramAddr+paramMinSizePos, param.minSize);
      ctx.arch.putInt(paramAddr+paramVarPosPos, param.varPos);
      ctx.arch.putInt(paramAddr+paramLocationPos, param.location);
    }
    return paramAddr;
  }
  
  private int genQualID(QualID qualID) {
    int qualIDAddr=0;
    if (qualID!=null) {
      if ((qualIDAddr=alloc(qualIDUnit))<0) return 0;
      ctx.arch.putInt(qualIDAddr+qualIDTypePos, qualID.type);
      ctx.arch.putRef(qualIDAddr-qualIDNamePos, genStringList(qualID.name));
      ctx.arch.putRef(qualIDAddr-qualIDPackDestPos, genPackage(qualID.packDest));
      ctx.arch.putRef(qualIDAddr-qualIDUnitDestPos, genUnit(qualID.unitDest, true));
    }
    return qualIDAddr;
  }
  
  private int genQualIDList(QualIDList qualIDList) {
    int qualIDListAddr=0;
    if (qualIDList!=null) {
      if ((qualIDListAddr=alloc(qualIDListUnit))<0) return 0;
      ctx.arch.putRef(qualIDListAddr-qualIDListQidPos, genQualID(qualIDList.qid));
      ctx.arch.putRef(qualIDListAddr-qualIDListNextPos, genQualIDList(qualIDList.nextQualID));
    }
    return qualIDListAddr;
  }
  
  private int genVrbl(Vrbl vrbl) {
    int vrblAddr=0;
    if (vrbl!=null) {
      if (vrbl.symHint!=0) return vrbl.symHint;
      if ((vrblAddr=alloc(vrblUnit))<0) return 0;
      ctx.arch.putRef(vrblAddr-vrblOwnerPos, genUnit(vrbl.owner, true));
      ctx.arch.putInt(vrblAddr+vrblModifierPos, vrbl.modifier);
      ctx.arch.putRef(vrblAddr-vrblNextVrblPos, genVrbl(vrbl.nextVrbl));
      
      ctx.arch.putRef(vrblAddr-vrblNamePos, ctx.allocateString(vrbl.name));
      ctx.arch.putRef(vrblAddr-vrblTypePos, genTypeRef(vrbl.type));
      ctx.arch.putInt(vrblAddr+vrblMinSizePos, vrbl.minSize);
      ctx.arch.putInt(vrblAddr+vrblVarPosPos, vrbl.varPos);
      ctx.arch.putInt(vrblAddr+vrblLocationPos, vrbl.location);
    }
    return vrblAddr;
  }
  
  private int genStringList(StringList stringList) {
    int stringListAddr=0;
    if (stringList!=null) {
      if ((stringListAddr=alloc(stringListUnit))<0) return 0;
      ctx.arch.putRef(stringListAddr-stringListStrPos, ctx.allocateString(stringList.str));
      ctx.arch.putRef(stringListAddr-stringListNextPos, genStringList(stringList.next));
      ctx.arch.putInt(stringListAddr+stringListTablePosPos, stringList.tablePos);
    }
    return stringListAddr;
  }
}
